from sqlalchemy.orm import sessionmaker
from typing import Dict, Any
from autolabel.transforms.schema import TransformCacheEntry
from autolabel.database import create_db_engine
from autolabel.data_models import Base
from typing import Optional
from autolabel.data_models import TransformCacheEntryModel
from .base import BaseCache
import logging

logger = logging.getLogger(__name__)


class SQLAlchemyTransformCache(BaseCache):
    """
    A cache system implemented with SQL Alchemy for storing the output of transforms.
    This cache system is used to avoid re-computing the output of transforms that have already been computed.
    This currently stores the input and the outputs of the transform.
    Caching is based on the transform name, params and input.
    """

    def __init__(self):
        self.engine = create_db_engine()
        self.base = Base
        self.session = None
        self.initialize()

    def initialize(self):
        self.base.metadata.create_all(self.engine)
        self.session = sessionmaker(bind=self.engine)()

    def lookup(self, entry: TransformCacheEntry) -> Optional[Dict[str, Any]]:
        """Retrieves an entry from the Cache. Returns None if not found.
        Args:
            entry: TransformCacheEntry we wish to retrieve from the Cache
        Returns:
            result: The output of the transform for this input. None if not found.
        """
        cache_entry = TransformCacheEntryModel.get(self.session, entry)
        if cache_entry is None:
            return None

        return cache_entry.output

    def update(self, entry: TransformCacheEntry) -> None:
        """Inserts the provided TransformCacheEntry into the Cache, overriding it if it already exists
        Args:
            entry: TransformCacheEntry we wish to put into the Cache
        """
        TransformCacheEntryModel.insert(self.session, entry)

    def clear(self, use_ttl: bool = True) -> None:
        """Clears the entire Cache based on ttl"""
        TransformCacheEntryModel.clear(self.session, use_ttl=use_ttl)



from typing import Dict, Any

from autolabel.transforms.schema import TransformType
from autolabel.transforms import BaseTransform
from autolabel.cache import BaseCache


class ImageTransform(BaseTransform):
    """This class is used to extract text from images using OCR. The output columns dictionary for this class should include the keys 'content_column' and 'metadata_column'

    This transform supports the following image formats: PNG, JPEG, TIFF, JPEG 2000, GIF, WebP, BMP, and PNM
    """

    COLUMN_NAMES = [
        "content_column",
        "metadata_column",
    ]

    def __init__(
        self,
        cache: BaseCache,
        output_columns: Dict[str, Any],
        file_path_column: str,
        lang: str = None,
    ) -> None:
        super().__init__(cache, output_columns)
        self.file_path_column = file_path_column
        self.lang = lang

        try:
            from PIL import Image
            import pytesseract

            self.Image = Image
            self.pytesseract = pytesseract
            self.pytesseract.get_tesseract_version()
        except ImportError:
            raise ImportError(
                "pillow and pytesseract are required to use the image transform with ocr. Please install pillow and pytesseract with the following command: pip install pillow pytesseract"
            )
        except EnvironmentError:
            raise EnvironmentError(
                "The tesseract engine is required to use the image transform with ocr. Please see https://tesseract-ocr.github.io/tessdoc/Installation.html for installation instructions."
            )

    @staticmethod
    def name() -> str:
        return TransformType.IMAGE

    def get_image_metadata(self, file_path: str):
        try:
            image = self.Image.open(file_path)
            metadata = {
                "format": image.format,
                "mode": image.mode,
                "size": image.size,
                "width": image.width,
                "height": image.height,
                "exif": image._getexif(),  # Exif metadata
            }
            return metadata
        except Exception as e:
            return {"error": str(e)}

    async def _apply(self, row: Dict[str, Any]) -> Dict[str, Any]:
        """This function transforms an image into text using OCR.

        Args:
            row (Dict[str, Any]): The row of data to be transformed.

        Returns:
            Dict[str, Any]: The dict of output columns.
        """
        content = self.pytesseract.image_to_string(
            row[self.file_path_column], lang=self.lang
        )
        metadata = self.get_image_metadata(row[self.file_path_column])
        transformed_row = {
            self.output_columns["content_column"]: content,
            self.output_columns["metadata_column"]: metadata,
        }
        return transformed_row

    def params(self) -> Dict[str, Any]:
        return {
            "output_columns": self.output_columns,
            "file_path_column": self.file_path_column,
            "lang": self.lang,
        }



from typing import List

from sklearn.metrics import classification_report

from autolabel.metrics import BaseMetric
from autolabel.schema import LLMAnnotation, MetricResult, MetricType


class ClassificationReportMetric(BaseMetric):
    def __init__(self) -> None:
        super().__init__()

    def compute(
        self, llm_labels: List[LLMAnnotation], gt_labels: List[str]
    ) -> List[MetricResult]:
        filtered_llm_labels = []
        filtered_gt_labels = []
        for llm_label, gt_label in zip(llm_labels, gt_labels):
            if llm_label.error is None and gt_label != "nan":
                filtered_llm_labels.append(llm_label)
                filtered_gt_labels.append(gt_label)

        filtered_llm_labels = [llm_label.label for llm_label in filtered_llm_labels]

        # if there are no labels, return empty list
        if len(filtered_gt_labels) == 0:
            return []

        # if the number of unique labels is too large, return empty list
        if len(set(filtered_gt_labels + filtered_llm_labels)) > 10:
            return []

        report = classification_report(filtered_gt_labels, filtered_llm_labels)

        value = [
            MetricResult(
                name=MetricType.CLASSIFICATION_REPORT,
                value=report,
                show_running=False,
            )
        ]
        return value



from abc import ABC, abstractmethod
from typing import List

from autolabel.schema import LLMAnnotation, MetricResult


class BaseMetric(ABC):
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def compute(
        self, llm_labels: List[LLMAnnotation], gt_labels: List[str]
    ) -> List[MetricResult]:
        pass



from .annotation import AnnotationModel
from .base import Base
from .dataset import DatasetModel
from .task import TaskModel
from .task_run import TaskRunModel
from .generation_cache import GenerationCacheEntryModel
from .transform_cache import TransformCacheEntryModel



import logging
from typing import Dict, List

from autolabel.configs import AutolabelConfig
from autolabel.schema import FewShotAlgorithm, ModelProvider
from langchain.embeddings import (
    CohereEmbeddings,
    HuggingFaceEmbeddings,
    OpenAIEmbeddings,
    VertexAIEmbeddings,
)
from langchain.embeddings.base import Embeddings
from langchain.prompts.example_selector import (
    MaxMarginalRelevanceExampleSelector,
    SemanticSimilarityExampleSelector,
)
from langchain.prompts.example_selector.base import BaseExampleSelector

from .fixed_example_selector import FixedExampleSelector
from .label_diversity_example_selector import (
    LabelDiversityRandomExampleSelector,
    LabelDiversitySimilarityExampleSelector,
)
from .vector_store import VectorStoreWrapper

ALGORITHM_TO_IMPLEMENTATION: Dict[FewShotAlgorithm, BaseExampleSelector] = {
    FewShotAlgorithm.FIXED: FixedExampleSelector,
    FewShotAlgorithm.SEMANTIC_SIMILARITY: SemanticSimilarityExampleSelector,
    FewShotAlgorithm.MAX_MARGINAL_RELEVANCE: MaxMarginalRelevanceExampleSelector,
    FewShotAlgorithm.LABEL_DIVERSITY_RANDOM: LabelDiversityRandomExampleSelector,
    FewShotAlgorithm.LABEL_DIVERSITY_SIMILARITY: LabelDiversitySimilarityExampleSelector,
}

DEFAULT_EMBEDDING_PROVIDER = OpenAIEmbeddings

PROVIDER_TO_MODEL: Dict[ModelProvider, Embeddings] = {
    ModelProvider.OPENAI: OpenAIEmbeddings,
    ModelProvider.GOOGLE: VertexAIEmbeddings,
    ModelProvider.HUGGINGFACE_PIPELINE: HuggingFaceEmbeddings,
    ModelProvider.COHERE: CohereEmbeddings,
}

logger = logging.getLogger(__name__)


class ExampleSelectorFactory:
    CANDIDATE_EXAMPLES_FACTOR = 5
    MAX_CANDIDATE_EXAMPLES = 100

    @staticmethod
    def initialize_selector(
        config: AutolabelConfig,
        examples: List[Dict],
        columns: List[str],
        cache: bool = True,
    ) -> BaseExampleSelector:
        algorithm = config.few_shot_algorithm()
        if not algorithm:
            return None
        try:
            algorithm = FewShotAlgorithm(algorithm)
        except ValueError as e:
            logger.error(
                f"{algorithm} is not in the list of supported few-shot algorithms: \
                {ALGORITHM_TO_IMPLEMENTATION.keys()}"
            )
            return None

        num_examples = config.few_shot_num_examples()
        params = config.vector_store_params()
        params["examples"] = examples
        params["k"] = num_examples
        if algorithm in [
            FewShotAlgorithm.SEMANTIC_SIMILARITY,
            FewShotAlgorithm.MAX_MARGINAL_RELEVANCE,
            FewShotAlgorithm.LABEL_DIVERSITY_SIMILARITY,
        ]:
            model_provider = config.embedding_provider()
            embedding_model_class = PROVIDER_TO_MODEL.get(
                model_provider, DEFAULT_EMBEDDING_PROVIDER
            )
            model_name = config.embedding_model_name()
            if model_name:
                embedding_model = embedding_model_class(model_name=model_name)
            else:
                embedding_model = embedding_model_class()
            params["embeddings"] = embedding_model
            params["vectorstore_cls"] = VectorStoreWrapper
            input_keys = [
                x
                for x in columns
                if x not in [config.label_column(), config.explanation_column()]
            ]
            params["input_keys"] = input_keys
        if algorithm == FewShotAlgorithm.MAX_MARGINAL_RELEVANCE:
            params["fetch_k"] = min(
                ExampleSelectorFactory.MAX_CANDIDATE_EXAMPLES,
                ExampleSelectorFactory.CANDIDATE_EXAMPLES_FACTOR * params["k"],
            )
        if algorithm in [
            FewShotAlgorithm.LABEL_DIVERSITY_RANDOM,
            FewShotAlgorithm.LABEL_DIVERSITY_SIMILARITY,
        ]:
            params["label_key"] = config.label_column()
            params["num_labels"] = len(config.labels_list())

        example_cls = ALGORITHM_TO_IMPLEMENTATION[algorithm]

        if algorithm not in [
            FewShotAlgorithm.FIXED,
            FewShotAlgorithm.LABEL_DIVERSITY_RANDOM,
        ]:
            params["cache"] = cache

        return example_cls.from_examples(**params)



from typing import List, Dict, Any

from autolabel.transforms.schema import TransformType
from autolabel.transforms import BaseTransform
from autolabel.cache import BaseCache


class PDFTransform(BaseTransform):
    """This class is used to extract text from PDFs. The output columns dictionary for this class should include the keys 'content_column' and 'metadata_column'"""

    COLUMN_NAMES = [
        "content_column",
        "metadata_column",
    ]

    def __init__(
        self,
        cache: BaseCache,
        output_columns: Dict[str, Any],
        file_path_column: str,
        ocr_enabled: bool = False,
        page_format: str = "Page {page_num}: {page_content}",
        page_sep: str = "\n\n",
        lang: str = None,
    ) -> None:
        super().__init__(cache, output_columns)
        self.file_path_column = file_path_column
        self.ocr_enabled = ocr_enabled
        self.page_format = page_format
        self.page_sep = page_sep
        self.lang = lang

        if self.ocr_enabled:
            try:
                from pdf2image import convert_from_path
                import pytesseract

                self.convert_from_path = convert_from_path
                self.pytesseract = pytesseract
                self.pytesseract.get_tesseract_version()
            except ImportError:
                raise ImportError(
                    "pdf2image and pytesseract are required to use the pdf transform with ocr. Please install pdf2image and pytesseract with the following command: pip install pdf2image pytesseract"
                )
            except EnvironmentError:
                raise EnvironmentError(
                    "The tesseract engine is required to use the pdf transform with ocr. Please see https://tesseract-ocr.github.io/tessdoc/Installation.html for installation instructions."
                )
        else:
            try:
                from langchain.document_loaders import PDFPlumberLoader

                self.PDFPlumberLoader = PDFPlumberLoader
            except ImportError:
                raise ImportError(
                    "pdfplumber is required to use the pdf transform. Please install pdfplumber with the following command: pip install pdfplumber"
                )

    @staticmethod
    def name() -> str:
        return TransformType.PDF

    def get_page_texts(self, row: Dict[str, Any]) -> List[str]:
        """This function gets the text from each page of a PDF file.
        If OCR is enabled, it uses the pdf2image library to convert the PDF into images and then uses
        pytesseract to convert the images into text. Otherwise, it uses pdfplumber to extract the text.

        Args:
            row (Dict[str, Any]): The row of data to be transformed.

        Returns:
            List[str]: A list of strings containing the text from each page of the PDF.
        """
        if self.ocr_enabled:
            pages = self.convert_from_path(row[self.file_path_column])
            return [
                self.pytesseract.image_to_string(page, lang=self.lang) for page in pages
            ]
        else:
            loader = self.PDFPlumberLoader(row[self.file_path_column])
            return [page.page_content for page in loader.load()]

    async def _apply(self, row: Dict[str, Any]) -> Dict[str, Any]:
        """This function transforms a PDF file into a string of text.
        The text is formatted according to the page_format and
        page_sep parameters and returned as a string.

        Args:
            row (Dict[str, Any]): The row of data to be transformed.

        Returns:
            Dict[str, Any]: The dict of output columns.
        """
        texts = []
        for idx, text in enumerate(self.get_page_texts(row)):
            texts.append(self.page_format.format(page_num=idx + 1, page_content=text))
        output = self.page_sep.join(texts)
        transformed_row = {
            self.output_columns["content_column"]: output,
            self.output_columns["metadata_column"]: {"num_pages": len(texts)},
        }
        return self._return_output_row(transformed_row)

    def params(self):
        return {
            "file_path_column": self.file_path_column,
            "ocr_enabled": self.ocr_enabled,
            "page_header": self.page_format,
            "page_sep": self.page_sep,
            "output_columns": self.output_columns,
            "lang": self.lang,
        }



from typing import List, Optional
import os

from autolabel.models import BaseModel
from autolabel.configs import AutolabelConfig
from autolabel.cache import BaseCache
from autolabel.schema import RefuelLLMResult


class CohereLLM(BaseModel):
    # Default parameters for OpenAILLM
    DEFAULT_MODEL = "command"
    DEFAULT_MODEL_PARAMS = {
        "max_tokens": 512,
        "temperature": 0.0,
    }

    # Reference: https://cohere.com/pricing
    COST_PER_TOKEN = 15 / 1_000_000

    def __init__(self, config: AutolabelConfig, cache: BaseCache = None) -> None:
        super().__init__(config, cache)
        try:
            import cohere
            from langchain.llms import Cohere
        except ImportError:
            raise ImportError(
                "cohere is required to use the cohere LLM. Please install it with the following command: pip install 'refuel-autolabel[cohere]'"
            )

        # populate model name
        self.model_name = config.model_name() or self.DEFAULT_MODEL

        # populate model params and initialize the LLM
        model_params = config.model_params()
        self.model_params = {
            **self.DEFAULT_MODEL_PARAMS,
            **model_params,
        }
        self.llm = Cohere(model=self.model_name, **self.model_params)
        self.co = cohere.Client(api_key=os.environ["COHERE_API_KEY"])

    def _label(self, prompts: List[str]) -> RefuelLLMResult:
        try:
            result = self.llm.generate(prompts)
            return RefuelLLMResult(
                generations=result.generations, errors=[None] * len(result.generations)
            )
        except Exception as e:
            return self._label_individually(prompts)

    def get_cost(self, prompt: str, label: Optional[str] = "") -> float:
        num_prompt_toks = len(self.co.tokenize(prompt).tokens)
        if label:
            num_label_toks = len(self.co.tokenize(label).tokens)
        else:
            num_label_toks = self.model_params["max_tokens"]

        return self.COST_PER_TOKEN * (num_prompt_toks + num_label_toks)

    def returns_token_probs(self) -> bool:
        return False



from collections import defaultdict
from typing import List, Dict, Tuple

from langchain.prompts.prompt import PromptTemplate

from autolabel.configs import AutolabelConfig
from autolabel.schema import LLMAnnotation, MetricType, MetricResult, F1Type
from autolabel.tasks import BaseTask
from autolabel.utils import get_format_variables

import json

from autolabel.metrics import (
    AccuracyMetric,
    AUROCMetric,
    SupportMetric,
    CompletionRateMetric,
    F1Metric,
    BaseMetric,
)


class MultilabelClassificationTask(BaseTask):
    DEFAULT_OUTPUT_GUIDELINES = 'You will return the answer as a semicolon-separated list of labels. For example: "label1;label2;label3"'
    DEFAULT_TASK_GUIDELINES = "Your job is to correctly label the provided input example into one or more of the following {num_labels} categories.\nCategories:\n{labels}\n"

    GENERATE_EXPLANATION_PROMPT = "You are an expert at providing a well reasoned explanation for the output of a given task. \n\nBEGIN TASK DESCRIPTION\n{task_guidelines}\nEND TASK DESCRIPTION\nYou will be given an input example and the corresponding output. Your job is to provide an explanation for why the output is correct for the task above.\nThink step by step and generate an explanation. The last line of the explanation should be - So, the answer is <label>.\n{labeled_example}\nExplanation: "

    def __init__(self, config: AutolabelConfig) -> None:
        super().__init__(config)
        self.metrics = [
            AccuracyMetric(),
            SupportMetric(),
            CompletionRateMetric(),
            F1Metric(
                type=F1Type.MULTI_LABEL,
                labels=self.config.labels_list(),
                sep=self.config.label_separator(),
                average=[MetricType.F1_MACRO, MetricType.F1_WEIGHTED],
            ),
        ]

        if self.config.confidence():
            self.metrics.append(AUROCMetric())

    def construct_prompt(self, input: Dict, examples: List) -> str:
        # Copy over the input so that we can modify it
        input = input.copy()

        # prepare task guideline
        labels_list = self.config.labels_list()
        num_labels = len(labels_list)
        fmt_task_guidelines = self.task_guidelines.format(
            num_labels=num_labels, labels="\n".join(labels_list)
        )

        # prepare seed examples
        example_template = self.config.example_template()
        label_column = self.config.label_column()
        fmt_examples = []
        for eg in examples:
            eg_copy = eg.copy()
            # If chain of thought is enabled
            if label_column and self.config.chain_of_thought():
                eg_copy[label_column] = json.dumps({label_column: eg[label_column]})
            fmt_examples.append(example_template.format_map(defaultdict(str, eg_copy)))

        # populate the current example in the prompt
        if label_column:
            input[label_column] = ""

        # populate the explanation column with empty string for current example
        explanation_column = self.config.explanation_column()
        if explanation_column:
            input[explanation_column] = ""

        # populate the current example in the prompt
        current_example = example_template.format_map(defaultdict(str, input))

        if self._is_few_shot_mode():
            return self.prompt_template.format(
                task_guidelines=fmt_task_guidelines,
                output_guidelines=self.output_guidelines,
                seed_examples="\n\n".join(fmt_examples),
                current_example=current_example,
            )
        else:
            return self.prompt_template.format(
                task_guidelines=fmt_task_guidelines,
                output_guidelines=self.output_guidelines,
                current_example=current_example,
            )

    def get_explanation_prompt(self, example: Dict) -> str:
        pt = PromptTemplate(
            input_variables=get_format_variables(self.GENERATE_EXPLANATION_PROMPT),
            template=self.GENERATE_EXPLANATION_PROMPT,
        )

        # prepare task guideline
        labels_list = self.config.labels_list()
        num_labels = len(labels_list)
        fmt_task_guidelines = self.task_guidelines.format(
            num_labels=num_labels, labels="\n".join(labels_list)
        )

        # prepare labeled example
        example_template = self.config.example_template()
        fmt_example = example_template.format_map(defaultdict(str, example))

        return pt.format(
            task_guidelines=fmt_task_guidelines,
            labeled_example=fmt_example,
        )

    def get_generate_dataset_prompt(
        self, label: str, num_rows: int, guidelines: str = None
    ) -> str:
        raise NotImplementedError("Dataset generation not implemented for this task")

    def eval(
        self,
        llm_labels: List[LLMAnnotation],
        gt_labels: List[str],
        additional_metrics: List[BaseMetric] = [],
    ) -> List[MetricResult]:
        """Evaluate the LLM generated labels by comparing them against ground truth

        Args:
            llm_labels (List[LLMAnnotation]): list of LLM generated labels
            gt_labels (List[str]): list of ground truth labels
            additional_metrics (List[BaseMetric], optional): list of additional metrics to compute. Defaults to [].

        Returns:
            List[MetricResult]: list of metrics and corresponding values
        """

        eval_metrics = []
        for metric in self.metrics + additional_metrics:
            eval_metrics.extend(metric.compute(llm_labels, gt_labels))

        return eval_metrics



from typing import Any, Optional

from sqlalchemy import create_engine
from sqlalchemy.engine.base import Engine

DB_ENGINE = None

# This creates one global ".autolabel.db" in your home directory.
# Having one global SQLite database in ~ is a poor idea, since
# SQLite cannot handle multiple simultaneous writes (if you have
# several different labeling jobs going on).
# DB_PATH = join(expanduser("~"), ".autolabel.db")

DB_PATH = ".autolabel.v15.db"


def create_db_engine(db_path: Optional[str] = DB_PATH) -> Engine:
    global DB_ENGINE
    if DB_ENGINE is None:
        DB_ENGINE = create_engine(f"sqlite:///{db_path}")
    return DB_ENGINE



from autolabel.cache import BaseCache
from autolabel.transforms import BaseTransform
from langchain.utilities import SerpAPIWrapper
from typing import Dict, Any
import logging
import pandas as pd

from autolabel.transforms.schema import (
    TransformError,
    TransformErrorType,
    TransformType,
)

logger = logging.getLogger(__name__)


class SerpApi(BaseTransform):
    COLUMN_NAMES = [
        "result_column",
    ]

    DEFAULT_ARGS = {
        "engine": "google",
        "google_domain": "google.com",
        "gl": "us",
        "hl": "en",
    }

    def __init__(
        self,
        cache: BaseCache,
        output_columns: Dict[str, Any],
        query_column: str,
        serp_api_key: str,
        serp_args: dict = DEFAULT_ARGS,
    ) -> None:
        super().__init__(cache, output_columns)
        self.query_column = query_column
        self.serp_api_key = serp_api_key
        self.serp_args = serp_args
        self.serp_api_wrapper = SerpAPIWrapper(
            search_engine=None, params=self.serp_args, serpapi_api_key=self.serp_api_key
        )

    def name(self) -> str:
        return TransformType.SERP_API

    async def _get_result(self, query):
        """
        Makes a request to Serp API with the query
        and returns the search results.
        """
        try:
            search_result = await self.serp_api_wrapper.arun(query=query)
        except Exception as e:
            raise TransformError(
                TransformErrorType.SERP_API_ERROR,
                f"Error while making request to Serp API: {e}",
            )
        return search_result

    async def _apply(self, row: Dict[str, Any]) -> Dict[str, Any]:
        query = row[self.query_column]
        search_result = self.NULL_TRANSFORM_TOKEN
        if pd.isna(query) or query == self.NULL_TRANSFORM_TOKEN:
            raise TransformError(
                TransformErrorType.INVALID_INPUT,
                f"Empty query in row {row}",
            )
        else:
            search_result = await self._get_result(query)

        transformed_row = {self.output_columns["result_column"]: search_result}

        return self._return_output_row(transformed_row)

    def params(self):
        return {
            "query_column": self.query_column,
            "output_columns": self.output_columns,
            "serp_api_key": self.serp_api_key,
            "serp_args": self.serp_args,
        }



"""Data and Schema Validation"""

import json
import re
from functools import cached_property
from json.decoder import JSONDecodeError
from typing import Dict, List, Optional, Union

from autolabel.configs import AutolabelConfig
from pydantic import BaseModel, ValidationError, create_model, root_validator
from pydantic.types import StrictStr

# Regex pattern to extract expected column from onfig.example_template()
EXPECTED_COLUMN_PATTERN = r"\{([^}]*)\}"


class NERTaskValidate(BaseModel):
    """Validate NER Task

    The label column can either be a string or a json
    """

    label_column: Optional[str]
    labels_set: Optional[set]  # A NER Task should have a unique set of labels in config

    def validate(self, value: str):
        """Validate NER

        A NER label can only be a dictionary
        """
        # TODO: This can be made better
        if value.startswith("{") and value.endswith("}"):
            try:
                seed_labels = json.loads(value)
                unmatched_label = set(seed_labels.keys()) - self.labels_set
                if len(unmatched_label) != 0:
                    raise ValueError(
                        f"labels: '{unmatched_label}' not in prompt/labels provided in config "
                    )
            except JSONDecodeError:
                raise
        else:
            raise


class ClassificationTaskValidate(BaseModel):
    """Validate Classification Task

    The label column can either be a string or a string of list
    """

    label_column: Optional[str]
    labels_set: Optional[
        set
    ]  # A classification Task should have a unique set of labels in config

    def validate(self, value: str):
        """Validate classification

        A classification label(ground_truth) could either be a list or string
        """
        # TODO: This can be made better
        if value.startswith("[") and value.endswith("]"):
            try:
                seed_labels = eval(value)
                if not isinstance(seed_labels, list):
                    raise
                unmatched_label = set(seed_labels) - self.labels_set
                if len(unmatched_label) != 0:
                    raise ValueError(
                        f"labels: '{unmatched_label}' not in prompt/labels provided in config "
                    )
            except SyntaxError:
                raise
        else:
            if value not in self.labels_set:
                raise ValueError(
                    f"labels: '{value}' not in prompt/labels provided in config "
                )


class EMTaskValidate(BaseModel):
    """Validate Entity Matching Task

    As of now we assume that the input label_column is a string
    """

    label_column: Optional[str]
    labels_set: Optional[
        set
    ]  # An EntityMatching Task should have a unique set of labels in config

    def validate(self, value: str):
        if value not in self.labels_set:
            raise ValueError(
                f"labels: '{value}' not in prompt/labels provided in config "
            )


class QATaskValidate(BaseModel):
    """Validate Question Answering Task

    As of now we assume that the input label_column is a string
    """

    label_column: Optional[str]
    labels_set: Optional[
        set
    ]  # A QA task may or may not have a unique set of label list

    def validate(self, value: str):
        """Since question answering is arbitarary task we have no validation"""
        pass


class MLCTaskValidate(BaseModel):
    """Validate Multilabel Classification Task

    As of now we assume that the input label_column is a string

    The label column can be a delimited string or a string of list
    """

    label_column: Optional[str]
    labels_set: Optional[
        set
    ]  # A Multilabel Classification Task should have a unique set of labels in config

    def validate(self, value: str):
        if value.startswith("[") and value.endswith("]"):
            try:
                seed_labels = eval(value)
                if not isinstance(seed_labels, list):
                    raise ValueError(
                        f"value: '{value}' is not a list of labels as expected"
                    )
                unmatched_label = set(seed_labels) - self.labels_set
                if len(unmatched_label) != 0:
                    raise ValueError(
                        f"labels: '{unmatched_label}' not in prompt/labels provided in config "
                    )
            except SyntaxError:
                raise
        else:
            # TODO: split by delimiter specified in config and validate each label
            pass


TaskTypeValidate = Union[
    NERTaskValidate,
    ClassificationTaskValidate,
    EMTaskValidate,
    QATaskValidate,
    MLCTaskValidate,
]


class DataValidationTasks(BaseModel):
    classification: TaskTypeValidate = ClassificationTaskValidate
    named_entity_recognition: TaskTypeValidate = NERTaskValidate
    entity_matching: TaskTypeValidate = EMTaskValidate
    question_answering: TaskTypeValidate = QATaskValidate
    multilabel_classification: TaskTypeValidate = MLCTaskValidate


class TaskDataValidation:
    """Task Validation"""

    def __init__(self, config: AutolabelConfig):
        """Task Validation

        Args:
            config: AutolabelConfig = User passed parsed configuration
        """
        # the type of task, classification, named_entity_recognition, etc.., "config/task_type"
        task_type: str = config.task_type()
        # the label column as specified in config, "config/dataset/label_column"
        label_column: str = config.label_column()
        # list of valid labels provided in config "config/prompt/labels"
        labels_list: Optional[List] = config.labels_list()

        # example template from config "config/prompt/example_template"
        self.example_template: str = config.example_template()
        # the explanation column as specified in config, "config/dataset/explanation_column"
        self.explanation_column: str = config.explanation_column()

        # the label column as specified in config, "config/dataset/label_column"
        self.label_column = config.label_column()

        self.__schema = {col: (StrictStr, ...) for col in self.expected_columns}

        self.__validation_task = DataValidationTasks.__dict__[task_type](
            label_column=label_column, labels_set=set(labels_list)
        )
        self.__data_validation = self.data_validation_and_schema_check(
            self.__validation_task
        )

    @cached_property
    def expected_columns(self) -> List:
        """Fetch expected columns"""
        column_name_lists = []
        for text in self.example_template.split("\n"):
            matches = re.findall(EXPECTED_COLUMN_PATTERN, text)
            column_name_lists += matches
        if self.explanation_column and self.explanation_column in column_name_lists:
            column_name_lists.remove(self.explanation_column)
        if self.label_column and self.label_column in column_name_lists:
            column_name_lists.remove(self.label_column)
        return column_name_lists

    @property
    def schema(self) -> Dict:
        """Fecth Schema"""
        return self.__schema

    @property
    def validation_task(
        self,
    ) -> TaskTypeValidate:
        """Fetch validation task"""
        return self.__validation_task

    def data_validation_and_schema_check(self, validation_task: BaseModel):
        """Validate data format and datatype

        Args:
            validation_task (TaskTypeValidate): validation task

        Raises:
            e: Validation error if the inputs are not string
            e: Validation error if validation_task fails

        Returns:
            DataValidation: Pydantic Model for validation
        """
        Model = create_model("Model", **self.__schema)

        class DataValidation(BaseModel):
            """Data Validation"""

            # We define validate as a classmethod such that a dynamic `data` can be passed
            # iteratively to the validate method using `DataValidation.validate`
            @classmethod
            def validate(cls, data):
                """Valdiate data types"""
                model = Model(**data)
                try:
                    # We perform the normal pydantic validation here
                    # This checks both the Schema and also calls check_fields
                    cls(**model.dict())
                except ValidationError as e:
                    raise e

            @root_validator(pre=True, allow_reuse=True)
            def check_fields(cls, values):
                """Validate data format"""
                try:
                    if (
                        validation_task.label_column
                        and validation_task.label_column in values
                    ):
                        label_column_value = values[validation_task.label_column]
                        validation_task.validate(label_column_value)
                except ValidationError as e:
                    raise e

        return DataValidation

    def validate(self, data: List[dict]) -> List[Dict]:
        """Validate Data"""
        error_messages = []
        for index, item in enumerate(data):
            try:
                self.__data_validation.validate(item)
            except ValidationError as e:
                for err in e.errors():
                    field = ".".join(err["loc"])
                    error_messages += [
                        {
                            "row_num": index,
                            "loc": field,
                            "msg": err["msg"],
                            "type": err["type"],
                        }
                    ]
        return error_messages

    def validate_dataset_columns(self, dataset_columns: List):
        """Validate columns

        Validate if the columns mentioned in example_template dataset are correct
        and are contained within the columns of the dataset(seed.csv)
        """
        missing_columns = set(self.expected_columns) - set(dataset_columns)
        assert (
            len(missing_columns) == 0
        ), f"columns={missing_columns} missing in seed.csv file"



from .base import Base
from sqlalchemy import Column, String, TEXT, Integer
import time
import pickle

from autolabel.transforms.schema import TransformCacheEntry


class TransformCacheEntryModel(Base):
    """an SQLAlchemy based Cache system for storing and retriving CacheEntries"""

    __tablename__ = "transform_cache"

    id = Column(String, primary_key=True)
    transform_name = Column(String(50))
    transform_params = Column(TEXT)
    input = Column(TEXT)
    output = Column(TEXT)
    creation_time_ms = Column(Integer)
    ttl_ms = Column(Integer)

    def __repr__(self):
        return f"<TransformCache(id={self.id},transform_name={self.transform_name},transform_params={self.transform_params},input={self.input},output={self.output})>"

    @classmethod
    def get(cls, db, cache_entry: TransformCacheEntry) -> TransformCacheEntry:
        id = cache_entry.get_id()
        looked_up_entry = db.query(cls).filter(cls.id == id).first()

        if not looked_up_entry:
            return None

        entry = TransformCacheEntry(
            transform_name=looked_up_entry.transform_name,
            transform_params=pickle.loads(looked_up_entry.transform_params),
            input=pickle.loads(looked_up_entry.input),
            output=pickle.loads(looked_up_entry.output),
            creation_time_ms=looked_up_entry.creation_time_ms,
            ttl_ms=looked_up_entry.ttl_ms,
        )
        return entry

    @classmethod
    def insert(cls, db, cache_entry: TransformCacheEntry) -> None:
        db_object = cls(
            id=cache_entry.get_id(),
            transform_name=cache_entry.transform_name,
            transform_params=pickle.dumps(cache_entry.transform_params),
            input=pickle.dumps(cache_entry.input),
            output=pickle.dumps(cache_entry.output),
            creation_time_ms=int(time.time() * 1000),
            ttl_ms=cache_entry.ttl_ms,
        )
        db.add(db_object)
        db.commit()
        return db_object

    @classmethod
    def clear(cls, db, use_ttl: bool = True) -> None:
        if use_ttl:
            current_time_ms = int(time.time() * 1000)
            db.query(cls).filter(
                current_time_ms - cls.creation_time_ms > cls.ttl_ms
            ).delete()
        else:
            db.query(cls).delete()
        db.commit()
